#include <avr/io.h>
#include <avr/interrupt.h>
#include "rossi.h"
//----------------------------------------------------------------------------------
//
// ROSSI-AVR 	RealTime Multitasking Operating System
// 7.2.2006 Heikki Palomki
//
/*
Modifications:
 18.2.2006	ATmega88 output compare register name
 25.8.2007	Enable task exchange in new task
 17.7.2009  ClrBox -function, tests with AT90USB647, header file structure


*/

// The normal values of definitions, defined in rossi.h
/*
#define Xtal 6000	// Clock frequency / kHz6MHz)
#define SysClock 1	// Scheduler period /ms
#define TaskNbr 5       // Scheduler period /ms
#define TaskSize 120    // Task stacksize /bytes
#define SigNbr 20       // Number of signals
#define TimerSign 100	// Time synchronization 1/100 sec
#define ResNbr 10       // Number of resources
#define BoxNbr 5        // Number of message boxes
#define BoxLen 20       // Size of box
#define Options 1       // Extra options
*/

unsigned char TaskStack[TaskNbr-1][TaskSize];//The stacks of other tasks than '0'
unsigned char TaskState[TaskNbr];        //States of task: 0=run, 1.2.3...=wait
void *TaskPointer[TaskNbr];              //Stack pointers
unsigned char TaskRun;                   //Number of running task
void *taskad;                             //Start address of new task code
unsigned int taskno=0;			 	 // and its number
#ifdef ResNbr
unsigned char Resources[ResNbr];         //Number of resources
#endif
#ifdef BoxNbr
unsigned int BoxS[BoxNbr],BoxP[BoxNbr],Boxes[BoxNbr][BoxLen+1]; //boxes
#endif
#ifdef Options
unsigned char Schedu;                    //Flag of visiting operating system ?
#endif
#ifdef SigNbr
#ifdef TimerSign
unsigned int sek;			 //Second counter
#endif
#endif

//------------------------------------------------------------------------------
//
// Start a routine from given address
//
void CallAddr(void (*Callrout)())
{ Callrout(); }

//------------------------------------------------------------------------------
//
// Initialization of ROSSI: task number =0, clear other tasks
//	clear boxes and resources, init timer0
//
void InitRossi(void)
 { 
 disable;
 int i;
 TaskRun = 0;                                 //this task running
 TaskState[0] = 0;                            //state is 'run'
 for(i=1; i<TaskNbr; i++) TaskState[i] = 255; //other task not usable
#if defined (ResNbr)
 for(i=0; i<ResNbr; i++) Resources[i] = 0;    //all resources free
#endif
#if defined (BoxNbr)
 for(i=0; i<BoxNbr; i++) {BoxS[i]=0; BoxP[i]=0;} //all boxes empty
#endif
 TCCR0 = 0b00000100;		        //TCNT0 speed: clk/256
 OCR0 = TCNT0 -1;			//3 ms time left
#if defined (TimerSign)
 TIMSK |= TOFON;			//enable overflow interrupt
#endif
 TIMSK |= TOCON;			//enable output compare interrupt
 enable;
 }

//------------------------------------------------------------------------------
//
// Next, ready-to-run task to run state, exchange context
//

// Executed by timer overflow of system function call
// or this executes the creation of new task
  
SIGNAL(OutputCompare)                   // Timer0 output compare
 {
 TIMSK &= TOCOFF;			//no interrupt overflow
#if defined (Options)
  Schedu = 1;				//task change flag ON
#endif
  asm("push R2");      //Other registers to stack
  asm("push R3");      //(compiler is stacking R0-R1, R18-R31)
  asm("push R4");
  asm("push R5");
  asm("push R6");
  asm("push R7");
  asm("push R8");
  asm("push R9");
  asm("push R10");
  asm("push R11");
  asm("push R12");
  asm("push R13");
  asm("push R14");
  asm("push R15");
  asm("push R16");
  asm("push R17");
  asm("push R28");	//stack pointer to stack (used in optimization O0)
  asm("push R29");
  int i,t;
//Here all task state date is stored in stack

//
//start of creating a new task
  if((taskno>0) & (taskno<TaskNbr))       //if number of task is inside allowed area
   {
   TaskPointer[taskno] = TaskStack[taskno]-1; //stack pointer to end of area
   TaskPointer[TaskRun] = (void *)SP;   //save present stack pointer
   SP = (int)TaskPointer[taskno];        //set new stack pointer
   TaskState[taskno]=0;                  //task ready to run
   TaskRun = taskno;
   taskno = 0;                           //set number of running (creating) task
   TIMSK |= TOCON;			//enable task exchange by output compare interrupt
   enable;
   CallAddr(taskad);                     //call new task
   TaskState[TaskRun]=255;		//task execution end
   }

  t=TaskRun+1;                          //seek next task
  if(t>=TaskNbr) t=0;                   //if last, restart seeking
  do                                    //seek next ready-to-run task
   {
   disable;
   for(i=t; (i<TaskNbr) & (TaskState[i] != 0); i++); //seek ready-to-run task
   if(i<TaskNbr)                        //if founded
     {
     TaskPointer[TaskRun] = (void *)SP; //save present stack pointer
     SP = (int)TaskPointer[i];  	//set new stack pointer
     TaskRun = i;               	//set the number of running task
     }
   t=0;					//if not foud, restart seeking
   OCR0 = TCNT0 + Xtal/256*SysClock;	//scheduler pediod in future
   TIFR = TOCON;			//clear possible interrupt flag
   enable;
   } while(i>=TaskNbr);         	//continue, until ready-to-run task is founded
  TIMSK |= TOCON;			//enable task scheduler again
  enable;                               
  asm("pop R29");			//load state (registers) from stack
  asm("pop R28");
  asm("pop R17");
  asm("pop R16");
  asm("pop R15");
  asm("pop R14");
  asm("pop R13");
  asm("pop R12");
  asm("pop R11");
  asm("pop R10");
  asm("pop R9");
  asm("pop R8");
  asm("pop R7");
  asm("pop R6");
  asm("pop R5");
  asm("pop R4");
  asm("pop R3");
  asm("pop R2"); 
  }
 


//------------------------------------------------------------------------------
//
//  Create a task with number nro and start it
//	Call formula: CreateTask(addr,num);
//
void CreateTask(void *tadr, int nro)
 {
 taskad=tadr; taskno=nro;     //store addres and number into global variables
 Schedule();                //so ccheduler can create new task
 }
 

#if defined (SigNbr)
//------------------------------------------------------------------------------
//
// Stay to wait signal 'nro'
//
void WaitSign(int nro)
 {
 TaskState[TaskRun] = nro;    //state = wait signal nro
 Schedule();                  //seek next task to run
 }

//------------------------------------------------------------------------------
//
// Send signal 'nro' e.g move waiting tasks to ready-to-run
//
void SendSign(int nro)
 {
 int s;
 for(s=0; s<TaskNbr; s++)     //seek all tasks
  {
  if(TaskState[s] == nro)     //if waiting, move to ready-to-run
     TaskState[s] = 0;
  }
 Schedule();                  //next task running
 }

//------------------------------------------------------------------------------
//
// Send signal 'nro' e.g move waiting tasks to ready-to-run
//
void PutSign(int nro)
 {
 disable;
 int s;
 for(s=0; s<TaskNbr; s++)    //seek all tasks
  {
  if(TaskState[s] == nro)    //if waiting, move to ready-to-run
     TaskState[s] = 0;
  }
 }	                     //disable stays !! , no schedule

#if defined (TimerSign)
//------------------------------------------------------------------------------
//
// Time syncronization send signal '1' after TimerSign/100 seconds
//
SIGNAL(TIMER0_OVF_vect
)
{
if(++sek>=(Xtal/197*3*TimerSign/100))
 {
 PutSign(1);                  //send signal without scheduler
 sek=0;
 }
}
#endif
#endif

#if defined (ResNbr)
//------------------------------------------------------------------------------
//
// Reserve resource 'nro' or stay waiting release
//
void ReserveRes(int nro)
 {
 disable;
 while(Resources[nro]>0)                 //if resource is reserved
  {
  TaskState[TaskRun] = nro+SigNbr;       //state= stay waiting this resource
  enable; Schedule();                    //next task running
  } 
 Resources[nro]++;                       //else, reserve resource
 enable;
 }

//------------------------------------------------------------------------------
//
// Release resource 'nro'
//
void FreeRes(int nro)
 {
 int s;
 disable;                                
  Resources[nro]=0;        		//release resource
 for(s=0; s<TaskNbr; s++)               //seek all tasks
  {
  if(TaskState[s] == (nro+SigNbr))      //if waiting this resource ..
     { TaskState[s] = 0; break; }       // move first of these to ready-to-run state
  }
 enable;
 // Schedule();                   	//next task running
 }

//------------------------------------------------------------------------------
//
// Ask reaource state
//
char AskRes(int nro)
 {
 if(Resources[nro]) return(0);
 else return(1);
 }
#endif

#if defined (BoxNbr)
//------------------------------------------------------------------------------
//
// Wait data from box 'nro'
//
int WaitBox(int nro)
 {
 int dat;
 disable;
 while(BoxS[nro]==BoxP[nro])                  //if empty ..
  {
  TaskState[TaskRun] = nro+SigNbr+ResNbr;     //state: stey waiting box
  enable;
  Schedule();                                 //next task running
  }
 dat=Boxes[nro][BoxS[nro]];                   //read data from box
 if(++BoxS[nro]>BoxLen) BoxS[nro]=0;          //circle buffer
 enable;
 return(dat);
 }

//------------------------------------------------------------------------------
//
// Ask box state
//
char AskBox(int nro)
 {
 if(BoxS[nro]==BoxP[nro]) return(0);	      //if empty, false
 else return(1);                              //else true
 }

//------------------------------------------------------------------------------
//
// Send data 'dat' to box 'nro'
//
char SendBox(int nro,int dat)
 {
 int PP,s;
 disable;
 if(BoxP[nro]>=BoxLen) PP=0; else PP=BoxP[nro]+1; //end pointer
 if(PP==(BoxS[nro]))		     	        //if box full ..
  {
  enable; 
  Schedule();				//next task running
  return(0);                            // and return false
  }
 Boxes[nro][BoxP[nro]] = dat;           //store data into buffer
 BoxP[nro] = PP;                        //end pointer to next free data
 for(s=0; s<TaskNbr; s++)               //seek all tasks
  {
  if(TaskState[s] == nro+SigNbr+ResNbr)	//if waiting box 'nro' ...
    {
	TaskState[s] = 0;               // move first to ready-to-run state
	break;
	}
  }
 //Schedule();                   	//next task running
 enable; return(1);                     //return true
 }

//------------------------------------------------------------------------------
//
// Send data 'dat' to box 'nro', interrupts disabled
//
char PutBox(int nro,int dat)
 {
 int PP,s;
 disable;
 if(BoxP[nro]>=BoxLen) PP=0; else PP=BoxP[nro]+1; //end pointer
 if(PP==(BoxS[nro])) return(0); 	//if box full, return false
 Boxes[nro][BoxP[nro]] = dat;           //store data into buffer
 BoxP[nro] = PP;                        //end pointer to next free data
 for(s=0; s<TaskNbr; s++)               //seek all tasks
  {
  if(TaskState[s] == nro+SigNbr+ResNbr) //if waiting box 'nro' ...
    TaskState[s] = 0;                   //move first to ready-to-run state
  }
 return(1);
 }

//------------------------------------------------------------------------------
//
// Empty box 'nro'
//
void ClrBox(int nro)
 {
 BoxS[nro]=0;
 BoxP[nro]=0;
 }

#endif

#if defined (Options)
//------------------------------------------------------------------------------
//
// Check, is task exchange executed
//
char Scheduled(void)
{
if(Schedu) {Schedu=0; return(1);}  //yes
else return(0);                    //no
}

//------------------------------------------------------------------------------
//
// Move task 'nro' to ready-to-run state
//
void RunTask(int nro)
 {
 TaskState[nro] = 0;
 }

//------------------------------------------------------------------------------
//
// End task 'nro' execution
//
void StopTask(int nro)
 {
 TaskState[nro] = 0xff;
 }

//------------------------------------------------------------------------------
//
// Ask task number
//
unsigned char MyTask(void)
 {
 return(TaskRun);
 }

#if defined (SysClock)
//------------------------------------------------------------------------------
//
// Disable scheduler
//
void Lock(void)
 {
 TIMSK &= TOCOFF;
 }

//------------------------------------------------------------------------------
//
// Enable sheduler
//
void Unlock(void)
 {
 TIMSK |= TOCON;
 }
#endif
#endif
